/******************************************************************************

   Hi-Q  --  by Tom Bajoras, written in Megamax C

******************************************************************************/

/*.............................. define things ..............................*/

   /* alert strings */
#define BADRSRC "[1][Unable to load the|resource file.][Sorry]"
#define CHEKQUIT  "[3][Are you SURE you|want to quit?][Cancel|Quit]"

#define RSRCNAME "HI-Q.RSC"       /* name of resource file */
#define NROWS 7                  /* # rows and # columns on playing board */
#define NCOLS 7
#define NROWS3 2                 /* # rows and # columns /3 */
#define NCOLS3 2
#define PEGPHASE 0               /* waiting for click on peg */
#define HOLEPHASE 1              /* waiting for click on hole */
#define MAXCOUNT 32              /* starting # of pegs */

#define ROOT 0                   /* starting level for AES object draw */
#define MAX_DEPTH 8              /* max depth for AES object draw */
#define SELECTED  0x1            /* object state bits */
#define OUTLINED 0x10

   /* object numbers from resource file */
#define MAINDIAL 0
#define HIQ_0_4 6
#define HIQ_1_4 7
#define HIQ_0_3 9
#define HIQ_0_2 8
#define HIQ_1_3 10
#define HIQ_1_2 11
#define HIQ_HELP 12
#define HIQ_2_0 14
#define HIQ_2_1 15
#define HIQ_4_0 17
#define HIQ_3_0 16
#define HIQ_3_1 18
#define HIQ_4_1 19
#define HIQ_2_2 21
#define HIQ_3_2 22
#define HIQ_4_2 23
#define HIQ_3_3 25
#define HIQ_2_3 24
#define HIQ_4_3 26
#define HIQ_2_4 27
#define HIQ_4_4 29
#define HIQ_3_4 28
#define HIQ_4_5 31
#define HIQ_3_5 33
#define HIQ_4_6 32
#define HIQ_2_5 34
#define HIQ_3_6 35
#define HIQ_2_6 36
#define HIQ_5_2 38
#define HIQ_5_3 39
#define HIQ_6_2 41
#define HIQ_5_4 40
#define HIQ_6_3 42
#define HIQ_6_4 43
#define HIQ_QUIT 46
#define HIQRESET 45
#define HIQCOUNT 44
#define HELPDIAL 1

/*........................ global variables .................................*/

int dummy;                    /* throw-away integer */
int gl_hand;                  /* VDI handle */
int *mainaddr,*helpaddr;      /* pointers to dialog box trees */
int dial_x,dial_y;            /* coordinates of centered dialog boxes */
int dial_w,dial_h;
int curr_r,curr_c;            /* currently selected peg (row #, column #) */
int count;                    /* current population (1-32) */
int lowest=MAXCOUNT;          /* best (lowest) score since boot-up */
int pegs[NROWS][NCOLS];       /* the population: 1 peg, 0 hole */
int exist[NROWS][NCOLS];      /* whether each peg is on the playing board */
int objects[NROWS][NCOLS] =   /* map gameboard to GEM objects, -1 for any
                                 that aren't on the playing board */
{ 
        -1,     -1,   HIQ_0_2,HIQ_0_3,HIQ_0_4,        -1,     -1,
        -1,     -1,   HIQ_1_2,HIQ_1_3,HIQ_1_4,        -1,     -1,
   HIQ_2_0,HIQ_2_1,   HIQ_2_2,HIQ_2_3,HIQ_2_4,   HIQ_2_5,HIQ_2_6,
   HIQ_3_0,HIQ_3_1,   HIQ_3_2,HIQ_3_3,HIQ_3_4,   HIQ_3_5,HIQ_3_6,
   HIQ_4_0,HIQ_4_1,   HIQ_4_2,HIQ_4_3,HIQ_4_4,   HIQ_4_5,HIQ_4_6,
        -1,     -1,   HIQ_5_2,HIQ_5_3,HIQ_5_4,        -1,     -1,
        -1,     -1,   HIQ_6_2,HIQ_6_3,HIQ_6_4,        -1,     -1,
};

/* communication with GEM */
int contrl[12], intin[128], ptsin[128], intout[128], ptsout[128];

/*................... get pointer to an object's state ......................*/

int
*ob_state(obj)
register int obj;      /* object number within main dialog tree */
{
   return( mainaddr + 12*obj + 5 );
}

/*................... get pointer to an object's width ......................*/

int
*ob_width(obj)
register int obj;      /* object number within main dialog tree */
{
   return( mainaddr + 12*obj + 10 );
}

/*................... get pointer to an object's height .....................*/

int
*ob_height(obj)
register int obj;      /* object number within tree */
{
   return( mainaddr + 12*obj + 11 );
}

/*...................... set bit in object's state word  ....................*/

set_obj(r,c,bit)
register int r,c,bit;
{
   *(ob_state(objects[r][c])) |= bit;
}

/*..................... clear bit in object's state word ....................*/

clr_obj(r,c,bit)
register int r,c,bit;
{
   *(ob_state(objects[r][c])) &= ~bit;
}

/*............................ main entry point .............................*/

main()
{
   if (startup()) exec();   /* if init'n succeeded, do body of program */
   finishup();              /* clean up before exiting */
}

/*............................... initialize ................................*/
/* return 1 for success, 0 for failure */

startup()      
{
   register int i,r,c;
   int loc_in[11],loc_out[57];

/* initialize AES */
   appl_init();

/* get VDI handle */
   gl_hand= graf_handle(&dummy,&dummy,&dummy,&dummy);

/* open VDI virtual workstation */
   for (i=0; i<10; i++) loc_in[i]=1;
   loc_in[10]= 2;
   v_opnvwk(loc_in,&gl_hand,loc_out);

/* arrow mouse */
   graf_mouse(0);

/* load resource file, abort if it can't be loaded */
   if ( !rsrc_load(RSRCNAME) ) 
   {
      form_alert(1,BADRSRC);
      return(0);
   }

/* get pointers to dialog box trees */
   rsrc_gaddr(0,MAINDIAL,&mainaddr);
   rsrc_gaddr(0,HELPDIAL,&helpaddr);

/* set up existence array:  each cell exists only if it is within the middle
   horizontal third or middle vertical third */
   for (r=0; r<NROWS; r++)
      for (c=0; c<NCOLS; c++)
      {
         exist[r][c]=1;
         if ( (r<NROWS3) || (r>(2*NROWS3)) )
            if ( (c<NCOLS3) || (c>(2*NCOLS3)) )
               exist[r][c]=0;
      }

/* get coordinates for centered dialog boxes (they are assumed to be exactly
   the same size) */
   form_center(mainaddr,&dial_x,&dial_y,&dial_w,&dial_h);
   form_center(helpaddr,&dummy,&dummy,&dummy,&dummy);

/* successfully initialized */
   return(1);
}

/*.......................... clean up and terminate .........................*/

finishup()
{
/* close VDI workstation */
   v_clsvwk(gl_hand);

/* clean up AES */
   appl_exit();
}

/*............................ body of program ..............................*/

exec()
{
   register int exitobj=HIQRESET;   /* reset game before first turn */
   register int phase;              /* alternates PEGPHASE/HOLEPHASE/... */
   register int r,c;
   int done;
   int mid_r,mid_c,mbuttons;

   do
   {
      done=0;                        /* default:  don't quit */

      switch (exitobj)
      {
         /* Quit button was clicked on */
         case HIQ_QUIT:
         
         done= form_alert(1,CHEKQUIT)-1;
         break;

         /* Reset button was clicked on */
         case HIQRESET:

         if (count<MAXCOUNT)     /* don't reset if already reset */
         {
            resetpegs();         /* all holes except center have pegs */
            curr_r=curr_c=(-1);  /* no currently selected peg */
            phase=PEGPHASE;      /* we're waiting for a peg to be clicked on */
            count=MAXCOUNT;      /* no pegs have been removed from board yet */
            resetobjs();         /* de-select all GEM objects */
            initscreen();        /* (re)draw the entire main dialog box */
         }
         break;

         /* Help button was clicked on:  display help dialog box */
         case HIQ_HELP:

         objc_draw(helpaddr,ROOT,MAX_DEPTH,dial_x,dial_y,dial_w,dial_h);
         form_do(helpaddr,-1);
         initscreen();            /* redraw the entire main dialog box */
         break;

         /* A peg or hole was clicked on */      
         default:

         /* Which peg or hole was clicked on? (find r and c) */
         for (r=0; r<NROWS; r++)
            for (c=0; c<NCOLS; c++)
               if (exitobj==objects[r][c]) goto found_it;
   
         found_it:      /* r and c have been defined */

         if (phase==PEGPHASE)    /* if waiting for a peg to be clicked on */
         {
            if (pegs[r][c])      /* if a peg (not a hole) was clicked on -- */
            {
               if ( can_it_jump(r,c) )    /* if this peg can jump  -- */
               {
                  curr_r=r;   curr_c=c;   /* make this peg the current peg */
                  set_obj(r,c,OUTLINED);  /* redraw this peg outlined */
                  draw_peg(r,c);
                  phase=HOLEPHASE;        /* now we're waiting for a hole */
               }
            }
         }
         else                    /* if waiting for a hole to be clicked on */
         {
            if ((r==curr_r)&&(c==curr_c))    /* if current peg re-clicked -- */
            {                                /* cancel the current peg */
               clr_obj(r,c,OUTLINED);        /* redraw it without outline */
               draw_peg(r,c);                /* wait for replacement for */
               phase=PEGPHASE;               /* current peg */
            }
            else
            {
               if (!pegs[r][c])      /* if a hole was clicked on -- */
               {
                  /* if current peg + this hole = a jump */
                  if (
                       ( (r==curr_r) && ( (c==(curr_c-2))||(c==(curr_c+2)) ) )
                     ||( (c==curr_c) && ( (r==(curr_r-2))||(r==(curr_r+2)) ) )
                      )
                  {
                     mid_r= (r+curr_r)/2;      /* which peg is to be jumped? */
                     mid_c= (c+curr_c)/2;   
                     if (pegs[mid_r][mid_c])
                     {
                        /* jump: update pegs and holes */
                        pegs[curr_r][curr_c]=0;
                        pegs[r][c]=1;
                        pegs[mid_r][mid_c]=0;
                        
                        /* jump:  update GEM object states */
                        clr_obj(curr_r,curr_c,SELECTED);
                        clr_obj(curr_r,curr_c,OUTLINED);
                        set_obj(r,c,SELECTED);
                        clr_obj(mid_r,mid_c,SELECTED);

                        /* jump:  redraw pegs and holes */                     
                        draw_peg(curr_r,curr_c);
                        draw_peg(mid_r,mid_c);
                        draw_peg(r,c);

                        /* a peg has been removed from the playing board */   
                        count--;
                        /* and possibly we've got a new best score */
                        if (count<lowest) lowest=count;
                        /* re-display count and lowest score */
                        showcount();

                        phase=PEGPHASE;      /* now we're waiting for a peg */
                     }
                  }
               }
            }               
         }
      }

      if (!done) 
      {
         exitobj= form_do(mainaddr,-1);      /* get touch exit */
         /* touch exits don't wait for button to be released, so
            to prevent buttons from auto-repeating we have to wait
            for buttons to clear */
         do
            graf_mkstate(&dummy,&dummy,&mbuttons,&dummy);
         while (mbuttons);
      }
   }
   while (!done);      /* loop until quit alert box is confirmed */
}

/*....................... can the given peg jump? ...........................*/

can_it_jump(r,c)      /* returns 1 or 0 (yes or no) */
register int r,c;
{
   register int jumped,jumped_to;

   /* There are four possible directions for jump:  up, down, left, or right.
      A jump can occur only if the following conditions are all true:
      
   (1) The cell that is to jump exists and has a peg in it.  (Assumed true.)
   (2) The cell that is to be jumped exists and has a peg in it.
   (3) The cell that is to be jumped to exists and does not have a peg in it.
   */

   /* jump up? */
   if (r>=NROWS3)
   {
      jumped=       pegs[r-1][c] && exist[r-1][c];
      jumped_to=   !pegs[r-2][c] && exist[r-2][c];
      if (jumped && jumped_to) return(1);
   }

   /* jump down? */
   if (r<=2*NROWS3)
   {
      jumped=      pegs[r+1][c] && exist[r+1][c];
      jumped_to=   !pegs[r+2][c] && exist[r+2][c];
      if (jumped && jumped_to) return(1);
   }      

   /* jump left? */
   if (c>=NCOLS3)
   {
      jumped=       pegs[r][c-1] && exist[r][c-1];
      jumped_to=   !pegs[r][c-2] && exist[r][c-2];
      if (jumped && jumped_to) return(1);
   }

   /* jump down? */
   if (c<=2*NCOLS3)
   {
      jumped=      pegs[r][c+1] && exist[r][c+1];
      jumped_to=   !pegs[r][c+2] && exist[r][c+2];
      if (jumped && jumped_to) return(1);
   }      

   return(0);
}

/*.............................. (re)init pegs ..............................*/

resetpegs()
{
   register int c,r;

   /* all existing cells have pegs -- */
   for (r=0; r<NROWS; r++)
      for (c=0; c<NCOLS; c++)
         if (exist[r][c])
            pegs[r][c]=1;
         else
            pegs[r][c]=0;

   /* -- except the center cell */
   pegs[NROWS/2][NCOLS/2]=0;
}

/*............................ reset all object state words .................*/

resetobjs()
{
   register int r,c;

   /* all pegs are selected, all holes are deselected */
   /* only the current peg is outlined */
   for (r=0; r<NROWS; r++)
   {
      for (c=0; c<NCOLS; c++)
      {
         if (exist[r][c])
         {
            if (pegs[r][c])
               set_obj(r,c,SELECTED);
            else
               clr_obj(r,c,SELECTED);
            if ((r==curr_r)&&(c==curr_r))
               set_obj(r,c,OUTLINED);
            else
               clr_obj(r,c,OUTLINED);
         }
      }
   }
}

/*............................. redraw entire screen ........................*/

initscreen()
{
/* draw the main dialog box from root down to maximum depth */
   objc_draw(mainaddr,ROOT,MAX_DEPTH,dial_x,dial_y,dial_w,dial_h);
   showcount();      /* show count and lowest score */
}

/*.......................... display count and lowest .......................*/

showcount()
{
   char charbuf[20],charbuf2[3];
   int x,y;

/* display count */
   strcpy(charbuf," Count: ");
   encode(count,charbuf2);
   strcat(charbuf,charbuf2);
   objc_offset(mainaddr,HIQCOUNT,&x,&y);
   v_gtext(gl_hand,x,y,charbuf);

/* display lowest */   
   strcpy(charbuf,"Lowest: ");
   encode(lowest,charbuf2);
   strcat(charbuf,charbuf2);
   v_gtext(gl_hand,x,y+16,charbuf);
}

/*...................... convert int to ascii decimal .......................*/

encode(number,buffer)
register int number;           /* number to be encoded, must be 0-99 */
register char *buffer;         /* 3 bytes (2 bytes + null terminator) */
{
   buffer[0]= '0' + number/10;
   buffer[1]= '0' + number%10;
   buffer[2]= 0;
}

/*......................... display a peg ...................................*/

draw_peg(r,c)
register int r,c;
{
   register int obj;
   int x,y,w,h;
   
/* convert row/column to GEM object number */
   obj=objects[r][c];

/* get position and size of this object */
   objc_offset(mainaddr,obj,&x,&y);
   w= *(ob_width(obj));
   h= *(ob_height(obj));

/* draw this object, slightly oversize the clipping rectangle so that the
   object's outline (if there is one) will be drawn */
   objc_draw(mainaddr,ROOT,MAX_DEPTH,x-4,y-4,w+8,h+8);
}

/************************************* EOF ***********************************/
